library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

library work;
use work.settings.all;
use work.test.all;

-- Testbench for comparator project.
--  * generates 48 MHz OSC_CLK and reset Signals (PO_RSTN, PB_RSTN)
--  * generates comparator inputs:
--    * VP = 1.5V +- 0.5V ramp, 5 kHz
--    * VN = 1.5V dc
--  * simulates the comparators non-ideal switching behavior
--  * applies various clock divider settings to the SWITCH input
--  * checks period and duty cycle of CLK and DCLK
--  * checks synchronization of RSTN and DRSTN to the corresponding clock signals
--  * checks LED, LOCK and COMPARE output
-- 
-- autor: Andreas Engel
-- date:  03/2013
entity TB is
end entity;
------------------------------------------------------------------------------------------------------------------------

architecture behav of TB is

  component TOP is
    port (
      OSC_CLK     : in  std_logic;
      PO_RSTN     : in  std_logic;
      PB_RSTN     : in  std_logic;
      VP          : in  std_logic;
      VN          : in  std_logic;
      SWITCH      : in  std_logic_vector(3 downto 0);
      LED         : out std_logic_vector(3 downto 0);
      LOCK        : out std_logic;
      CLK         : out std_logic;
      RSTN        : out std_logic;
      DCLK        : out std_logic;
      DRSTN       : out std_logic;
      COMPARE     : out std_logic);
  end component;
  
  component OPAMP_SIM is
    port (
      VPR : in  real;
      VNR : in  real;
      VP  : out std_logic;
      VN  : out std_logic);
  end component;
  
  signal s_osc_clk     : std_logic := '0';
  signal s_po_rstN     : std_logic := '0';
  signal s_pb_rstN     : std_logic := '0';
  signal s_vp          : std_logic;
  signal s_vn          : std_logic;
  signal s_div         : positive;
  signal s_switch      : std_logic_vector(3 downto 0);
  signal s_led         : std_logic_vector(3 downto 0);
  signal s_lock        : std_logic;
  signal s_clk         : std_logic;
  signal s_rstN        : std_logic;
  signal s_dclk        : std_logic;
  signal s_drstN       : std_logic;
  signal s_compare     : std_logic;
  signal s_dclk_assert : boolean;
  signal s_clk_cycles  : natural := 0;
  signal s_dclk_cycles : natural := 0;
  
  signal vpr           : real := 1.0;
  signal vnr           : real := 1.5;
begin

  -- clock and reset generation
  s_osc_clk <= not s_osc_clk after 0.5 us / 48;
  s_po_rstN <= '1'           after 0.2 us / 48;
  s_pb_rstN <= '1'           after 0.4 us / 48;
  
  -- generate positive comparator input
  vpr <= vpr + 0.0001 after 10 ns when (integer(now / 1 us) mod 200) < 100 else
         vpr - 0.0001 after 10 ns;
  
  -- simulate comparator behavoir
  i_opamp: OPAMP_SIM
    port map (
      VPR => vpr,
      VNR => vnr,
      VP  => s_vp,
      VN  => s_vn);
  
  -- each divider selected for 6 dclk cycles
  -- (disable period/duty cycle assertions for first rising edege after divider change)
  p_divider: process
  begin
    s_div         <= 1;
    s_dclk_assert <= false;
    wait until rising_edge(s_lock);
    for k in 1 to 2**s_switch'length loop
      s_div <= k;
      wait until falling_edge(s_dclk);
      s_dclk_assert <= true;
      for i in 1 to 5 loop
        wait until falling_edge(s_dclk);
      end loop;
      s_dclk_assert <= false;
    end loop;
    wait;
  end process;
  s_switch <= std_logic_vector(to_unsigned(s_div-1,s_switch'length));
  
  -- unit under test
  i_uut: TOP
    port map (
      OSC_CLK     => s_osc_clk,
      PO_RSTN     => s_po_rstN,
      PB_RSTN     => s_pb_rstN,
      VP          => s_vp,
      VN          => s_vn,
      SWITCH      => s_switch,
      LED         => s_led,
      LOCK        => s_lock,
      CLK         => s_clk,
      RSTN        => s_rstN,
      DCLK        => s_dclk,
      DRSTN       => s_drstN,
      COMPARE     => s_compare);
  
  -- check period and duty cycle of clk signal
  p_clk: process (s_clk)
    variable v_rising      : time;
    variable v_falling     : time;
    variable v_half_period : time;
    variable v_period      : time;
  begin
    if rising_edge(s_clk) then
      v_half_period := now - v_falling;
      v_period      := now - v_rising;
      v_rising      := now;
      s_clk_cycles  <= s_clk_cycles + 1;
    else
      v_half_period := now - v_rising;
      v_period      := now - v_falling;
      v_falling     := now;
    end if;
    if s_rstN = '1' then
      test(v_half_period > 0.45 * 1 us / 4.0 and v_half_period < 0.55 * 1 us / 4.0, 
           "unexpected CLK duty cycle " & real'image(v_half_period / 1 ns * 0.004 * 100.0));
      test(v_period      > 0.95 * 1 us / 4.0 and v_period      < 1.05 * 1 us / 4.0, 
           "unexpected CLK period " & time'image(v_period));
    end if;
  end process;
  
  -- check period and duty cycle of dclk signal
  p_dclk: process (s_dclk)
    variable v_rising      : time;
    variable v_falling     : time;
    variable v_half_period : time;
    variable v_period      : time;
  begin
    if rising_edge(s_dclk) then
      v_half_period := now - v_falling;
      v_period      := now - v_rising;
      v_rising      := now;
      s_dclk_cycles <= s_dclk_cycles + 1;
    else
      v_half_period := now - v_rising;
      v_period      := now - v_falling;
      v_falling     := now;
    end if;
    if s_dclk_assert then
      test(v_half_period > 0.32 * 1 us / 4.0 * s_div and v_half_period < 0.67 * 1 us / 4.0 * s_div, 
           "unexpected DCLK duty cycle " & real'image(v_half_period / 1 ns * 0.004 / real(s_div) * 100.0)); 
      test(v_period      > 0.95 * 1 us / 4.0 * s_div and v_period      < 1.05 * 1 us / 4.0 * s_div,
           "unexpected DCLK period " & time'image(v_period));
    end if;
  end process;
  
  p_cycles: process
  begin
    wait for 199 us;
    test(s_clk_cycles  > 790, "too few CLK cylces detected");
    test(s_dclk_cycles >  90, "too few DCLK cylces detected");
  end process;
  
  -- check rstN signal
  p_rstN: process
  begin
    wait for 30 ns;   -- IO delay (postlayout)
    test(s_rstN, '0', "RSTN");
    wait until rising_edge(s_clk);
    wait for 10 ns;   -- IO delay (postlayout)
    test(s_rstN, '0', "RSTN");
    wait until rising_edge(s_clk);
    wait until rising_edge(s_clk);
    wait for 10 ns;   -- IO delay (postlayout)
    test(s_rstN, '1', "RSTN");
    
    wait; 
  end process;
  
  -- check drstN signal
  p_drstN: process
  begin
    wait for 35 ns;   -- IO delay (postlayout)
    test(s_drstN, '0', "DRSTN");
    wait until rising_edge(s_dclk);
    wait for 10 ns;   -- IO delay (postlayout)
    test(s_drstN, '0', "DRSTN");
    wait until rising_edge(s_dclk);
    wait until rising_edge(s_dclk);
    wait for 10 ns;   -- IO delay (postlayout)
    test(s_drstN, '1', "DRSTN");
    wait; 
  end process;
  
  -- check LED output
  p_led: process
  begin
    wait until s_led'event or s_switch'event;
    wait for 19 ns;   -- IO delay (postlayout)
    test (s_led = s_switch, "unexpected LED output"); 
  end process;
  
  -- check LOCK signal
  p_lock: process
  begin
    wait until s_po_rstN'event or s_pb_rstN'event or s_lock'event;
    wait for 30 ns;   -- IO delay (postlayout)
    if (s_po_rstN and s_pb_rstN) = '0' then
      test(s_lock, '0', "LOCK");
    end if;
  end process;
  
  -- check COMPARE signal (may only change after DCLK rising)
  p_compare: process
  begin
    wait until s_compare'event;
    if (s_drstN = '1') then
      wait for 1 ns;
      test(s_dclk = '1' and s_dclk'last_event < 10 ns, "COMPARE switched unexpected");
    end if;
  end process;
  
end architecture; 
