library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

library work;
use work.settings.all;
use work.test.all;

-- Simulation of PGA behavior.
--  * receive serial configuration data
--  * check timing of serial interface
--  * calculate VOUT = VREF + (VIN-VREF) * gain
-- 
-- author: Andreas Engel
-- date:   03/13
entity PGA_SIM is
  port (
    SCK  : in  std_logic;        -- SPI clock
    SDI  : in  std_logic;        -- SPI data
    CSN  : in  std_logic;        -- SPI chip select (active low)
    VIN  : in  real;             -- input voltage (to be scaled)
    VOUT : out real);            -- scaled output voltage
end entity;
------------------------------------------------------------------------------------------------------------------------


architecture beh of PGA_SIM is
  signal s_shift : std_logic_vector(7 downto 0)    := (others => '0');
  signal s_hold  : std_logic_vector(s_shift'range);
begin
  -- serial shift in
  p_conf: process(SCK)
  begin
    if rising_edge(SCK) and CSN = '0' then
      s_shift <= s_shift(s_shift'high-1 downto 0) & SDI;
    end if;
  end process;

  -- hold register (double buffered configuration)
  p_hold: process(CSN)
  begin
    if rising_edge(CSN) and SCK = '1' then
      s_hold <= s_shift;
    end if;
  end process;

  -- output stage
  VOUT <= 0.0  when s_hold(4) = '1' else                                        -- power down
          VREF when s_hold(5) = '1' else                                        -- input zeroing
          VREF + (VIN-VREF) * real(to_integer(unsigned(s_hold(3 downto 0))+1)); -- amplificaton
  
  -- Serial Interface Timing requirements (see page 7)
  p_timing : process(SCK, SDI, CSN) is
    type t_timing is array(1 to 11) of time; 
    constant t : t_timing := (
      1   => 100 ns,   -- min sck high
      2   => 100 ns,   -- min sck low
      3   =>  50 ns,   -- min csN falling to sck falling
      4   =>  30 ns,   -- min sdi valid before sck rising
      5   =>  10 ns,   -- min sdi hold after sclk rising
      6   =>  60 ns,   -- max sck falling to sdo valid
      7   =>  50 ns,   -- min sck transition to csN rising
      8   =>  50 ns,   -- min csN high
      9   =>  50 ns,   -- max csN falling to sdo active
     10   =>  50 ns,   -- max csN rising to sdo inactive
     11   =>  10 ns);  -- min sck transition to csN falling
    
    variable v_sck_rising  : time := 0 ns;
    variable v_sck_falling : time := 0 ns;
    variable v_csN_rising  : time := 0 ns;
    variable v_csN_falling : time := 0 ns;
    
    procedure test_timing(n : in natural; b : in boolean) is 
    begin 
      test(b, "PGA timing t" & natural'image(n) & " violated");
    end procedure; 
  begin
  
    -- update reference points
    if rising_edge(SCK)  then v_sck_rising  := now; end if;
    if falling_edge(SCK) then v_sck_falling := now; end if;
    if rising_edge(CSN)  then v_csN_rising  := now; end if;
    if falling_edge(CSN) then v_csN_falling := now; end if;
  
    -- no tests before reset completed
    if now > 3 us then
      a_1: if falling_edge(SCK) then test_timing(1,  now - v_sck_rising  >= t(1));  end if;
      a_2: if rising_edge(SCK)  then test_timing(2,  now - v_sck_falling >= t(2));  end if;
      a_3: if falling_edge(SCK) then test_timing(3,  CSN'stable(t(3)));             end if;
      a_4: if rising_edge(SCK)  then test_timing(4,  SDI'stable(t(4)));             end if;
      a_5: if SDI'event         then test_timing(5,  now - v_sck_rising  >= t(5));  end if;
      a_7: if rising_edge(CSN)  then test_timing(7,  SCK='1' and SCK'stable(t(7))); end if;
      a_8: if falling_edge(CSN) then test_timing(8,  now - v_csN_rising  >= t(8));  end if;
      a_11:if falling_edge(CSN) then test_timing(11, SCK='1' and SCK'stable(t(11)));end if;
    end if;
  end process;

end architecture;
