library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use IEEE.math_real.all;

library work;
use work.settings.all;
use work.test.all;


-- Testbench for system project.
--  * generates 48 MHz OSC_CLK and reset Signals (PO_RSTN, PB_RSTN)
--  * generates input voltage (1.15V, 1.35V)
--  * generate  gain settings (1, MAX_GAIN)
--  * simulate PGA behavior
--  * simulate comparator behavior
--  * simulate charge/discharge of Vcompare
--  * receive samples via UART
--  * calculate min / max / rms error of sampled values
-- 
-- autor: Andreas Engel
-- date:  03/2013
entity tb is
end tb;
------------------------------------------------------------------------------------------------------------------------


architecture behavior of TB is

  component TOP is
    port (
      OSC_CLK  : in  std_logic;
      PO_RSTN  : in  std_logic;
      PB_RSTN  : in  std_logic;
      SWITCH   : in  std_logic_vector(4 downto 0);
      LED      : out std_logic_vector(4 downto 0);
      VSCALED  : in  std_logic;
      VCOMPARE : in  std_logic;
      VCHARGE  : out std_logic;
      TX       : out std_logic;
      SCK      : out std_logic;
      SDI      : out std_logic;
      CSN      : out std_logic);
  end component;
  
  component PGA_SIM is
    port (
      SCK      : in  std_logic;
      SDI      : in  std_logic;
      CSN      : in  std_logic;
      VIN      : in  real;
      VOUT     : out real);
  end component;

  component OPAMP_SIM is
    port (
      VPR      : in  real;
      VNR      : in  real;
      VP       : out std_logic;
      VN       : out std_logic);
  end component;
  
  signal s_osc_clk   : std_logic := '0';
  signal s_pb_rstN   : std_logic := '1';
  signal s_po_rstN   : std_logic := '0';
     
  signal vin         : real;
  signal vscaled     : real;
  signal vcompare    : real := VCC;

  signal s_scaled    : std_logic;
  signal s_compare   : std_logic;
  signal s_charge    : std_logic;

  signal s_csN       : std_logic;
  signal s_sck       : std_logic;
  signal s_sdi       : std_logic;

  signal s_switch    : std_logic_vector(4 downto 0);
  signal s_led       : std_logic_vector(4 downto 0);
  signal s_rx        : std_logic;
  
  signal s_gain      : natural   := 1;
  signal s_calibrate : std_logic := '0'; 
  
  signal result      : real;
  signal results     : natural :=    0;
  signal error       : real    :=    0.0;
  signal rms         : real    :=    0.0;
  signal max         : real    := -100.0;
  signal min         : real    :=  100.0;
  
  signal end_of_sim  : std_logic := '0';
begin

  -- clock and reset generation
  s_osc_clk <= not s_osc_clk after 0.5 us / 48;
  s_po_rstN <= '1'           after   5 us / 48;
  
  -- RC-model
  p_rc: process
    variable v_vcharge : real;
  begin
    if s_charge = '1' then v_vcharge := VCC;
                      else v_vcharge := 0.0; end if;
    vcompare <= vcompare + (v_vcharge - vcompare) * (1.0 - exp(- OPAMP_UPDATE / 1000.0 / real(TAU)));
    wait for 1 ns * OPAMP_UPDATE;
  end process;
  
  -- simulate PGA behavior
  i_pga: PGA_SIM
    port map (
      SCK  => s_sck,
      SDI  => s_sdi,
      CSN  => s_csN,
      VIN  => vin,
      VOUT => vscaled);
  
  -- simulate comparator behavoir
  i_opamp: OPAMP_SIM
    port map (
      VPR  => vcompare,
      VNR  => vscaled,
      VP   => s_compare,
      VN   => s_scaled);
  
  -- combine user input to switch signal
  s_switch(3 downto 0) <= std_logic_vector(to_unsigned(s_gain-1, 4));
  s_switch(4)          <= s_calibrate;
  
  -- unit under test
  i_uut: TOP
    port map (
      OSC_CLK  => s_osc_clk,
      PB_RSTN  => s_pb_rstN,
      PO_RSTN  => s_po_rstN,
      SWITCH   => s_switch,
      LED      => s_led,
      TX       => s_rx,
      VSCALED  => s_scaled,
      VCOMPARE => s_compare,
      VCHARGE  => s_charge,
      SDI      => s_sdi,
      SCK      => s_sck,
      CSN      => s_csN);
  
  -- serial data reception and result calculation
  b_urx: block is
    component URX is
      port(
        RX       : in  std_logic;
        VALID    : out std_logic;
        BYTE     : out std_logic_vector(7 downto 0));
    end component;
    signal s_byte  : std_logic_vector(7 downto 0);
    signal s_valid : std_logic;
  begin
    i_byte: URX
      port map (
        RX    => s_rx,
        VALID => s_valid,
        BYTE  => s_byte);

    p_acc: process(s_valid, s_calibrate)
      constant c_bytes : positive := (SAMPLE_WIDTH + 4 + 7) / 8;
      variable v_byte  : natural  := 0;
      variable v_acc   : natural  := 0;
    begin
      if s_calibrate = '1' then
        results  <= 0;
      elsif rising_edge(s_valid) then
        v_acc  := v_acc + to_integer(unsigned(s_byte)) * 256**v_byte;
        v_byte := v_byte + 1;
      
        if v_byte = c_bytes then    -- all bytes accumulated => calculate result
          result  <= VCC * real(v_acc) / 2.0**real(SAMPLE_WIDTH+4);
          results <= results + 1;
          v_acc     := 0;
          v_byte    := 0;
        end if;
      end if;
    end process;
  end block;
  
  -- user interface: set input voltage, calibrate and gain
  p_ui: process
  begin
    s_calibrate  <= '0';
    vin          <= VREF+0.1;
    s_gain       <= 1;
    
    -- wait until PLL locked
    wait for 10 us;
    
    -- sweep gain (in pseudo random order, to test MAX_GAIN saturation)
    for g in 1 to 16 loop
      s_gain <= g/2 + 8*(g mod 2);
      wait for 5 us;
    end loop;
    wait for 5 us;
    
    report "Test sampling error for SAMPLE_WIDTH = " & integer'image(SAMPLE_WIDTH) & 
                                   " RC_ACCURACY = " & integer'image(RC_ACCURACY);
    
    -- check measurement accuracy for certain gains and input voltages
    for g in 0 to 1 loop
      s_gain <= 1+g*(MAX_GAIN-1);
      for i in 0 to 1 loop
        vin <= 1.15 + real(i)*0.2;
        wait until results = 5;
        
        wait for 10 us;
        report " Vin  = "         & real'image(vin)       & 
               " Gain = "         & integer'image(s_gain) & 
               " => rms error = " & real'image(rms)       & " %";
        
        s_calibrate <= '1';
        wait for 4 us;
        s_calibrate <= '0';
      end loop;
    end loop;
    
    -- stop simulation
    wait for 100 us;
    report "min/max error = " & real'image(min) & " % / " & real'image(max) & " %";
    end_of_sim <= '1';
    
    wait;
  end process;
  
  -- check GAIN configuration of PGA
  p_gain: process
  begin
    wait until s_gain'event;
    wait for 3.0 us; -- SPI config time
    if s_gain < MAX_GAIN then test(vscaled, VREF + (vin-VREF)*real(s_gain),   "VSCALED");
                         else test(vscaled, VREF + (vin-VREF)*real(MAX_GAIN), "VSCALED"); end if;
  end process;
  
  -- calculate result error + local rms (for each GAIN/VIN combination) + global min/max error
  p_error: process(results)
    variable v_error : real;
    variable v_cnt   : natural := 0;
    variable v_acc   : real    := 0.0;
  begin
    if (results > 0) then
      v_error := (result-vin) / vin * 100.0;
      error <= v_error;
      test(abs(v_error) <= 10.0, "Measurement error exceeds 10%");

      if (results = 1) then v_acc := 0.0; end if;
      v_acc := v_acc + v_error * v_error;
      rms <= sqrt(v_acc/real(results));

      if (v_error > max) then max <= v_error; end if;
      if (v_error < min) then min <= v_error; end if;
    end if;
  end process;

end architecture; 
