`default_nettype none
`timescale 1ns/100ps

// Simulation of PGA behavior.
//  * receive serial configuration data
//  * check timing of serial interface
//  * calculate VOUT = VREF + (VIN-VREF) * gain
// 
// author: Andreas Engel
// date:   03/13
module PGA_SIM #(`include "settings.v") 
 (input  wire        SCK,              // SPI clock                   
  input  wire        SDI,              // SPI data                    
  input  wire        CSN,              // SPI chip select (active low)
  input  wire [63:0] VIN,              // input voltage (to be scaled)
  output wire [63:0] VOUT);            // scaled output voltage       
/**********************************************************************************************************************/

// serial shift in
reg [7:0] shift = 8'd0;
always @(posedge SCK) if (~CSN) shift <= {shift[6:0], SDI};

// hold register (double buffered configuration)
reg [7:0] hold = 8'd0;
always @(posedge CSN) if (SCK) hold <= shift;

// output stage
assign VOUT = $realtobits (
  hold[4] ? 0.0  :                                 // power down
  hold[5] ? VREF :                                 // input zeroing
  VREF + ($bitstoreal(VIN)-VREF) * (hold[3:0]+1)   // amplificaton
); 
  
// Serial Interface Timing requirements (see page 7) in [ns]
localparam  t1 = 100; // min sck high
localparam  t2 = 100; // min sck low
localparam  t3 =  50; // min csN falling to sck falling
localparam  t4 =  30; // min sdi valid before sck rising
localparam  t5 =  10; // min sdi hold after sclk rising
localparam  t6 =  60; // max sck falling to sdo valid
localparam  t7 =  50; // min sck transition to csN rising
localparam  t8 =  50; // min csN high
localparam  t9 =  50; // max csN falling to sdo active
localparam t10 =  50; // max csN rising to sdo inactive
localparam t11 =  10; // min sck transition to csN falling

// check timing
`include "test.v"
task testT;
  input [3:0] n;
  input  b;
  reg   [8*80-1:0] m;
  // no tests before reset completed
  if ($realtime > 3000.0) begin
    $sformat(m, "PGA timing t%-0d violated", n);
    testM(b, m);
  end
endtask

real sck_rising  = 0;
real sck_falling = 0;
real csN_rising  = 0;
real csN_falling = 0;
real sck_event   = 0;
real csN_event   = 0;
real sdi_event   = 0;

always @(posedge SCK) sck_rising  = $realtime;
always @(negedge SCK) sck_falling = $realtime;
always @(posedge CSN) csN_rising  = $realtime;
always @(negedge CSN) csN_falling = $realtime;
always @(SCK)         sck_event   = $realtime;
always @(CSN)         csN_event   = $realtime;
always @(SDI)         sdi_event   = $realtime;

always @(negedge SCK) testT(1,  $realtime - sck_rising  >= t1);
always @(posedge SCK) testT(2,  $realtime - sck_falling >= t2);
always @(negedge SCK) testT(3,  $realtime - csN_event   >= t3);                 
always @(posedge SCK) testT(4,  $realtime - sdi_event   >= t4);                 
always @(SDI)         testT(5,  $realtime - sck_rising  >= t5);
always @(posedge CSN) testT(7,  $realtime - sck_event   >= t7  && SCK == 1'b1);     
always @(negedge CSN) testT(8,  $realtime - csN_rising  >= t8);
always @(negedge CSN) testT(11, $realtime - sck_event   >= t11 && SCK == 1'b1);    

endmodule
